from typing import Any


def add_list_arguments(parser: Any) -> None:
    add_list_flows_arguments(parser)
    add_list_keys_arguments(parser)
    add_list_models_arguments(parser)
    add_list_wrappers_arguments(parser)

def add_list_wrappers_arguments(parser: Any) -> None:
    parser.add_parser("wrappers", help="list supported wrappers")

def add_list_models_arguments(parser: Any) -> None:
    parser.add_parser("models", help="list existing models")

def add_list_keys_arguments(parser: Any) -> None:
    parser.add_parser("keys", help="list existing keys")

def add_list_flows_arguments(parser: Any) -> None:
    parser.add_parser("flows", help="list existing flows")



from fastapi import APIRouter
# from omnibridge.api.app.api_v1.endpoints import hello_world
from omnibridge_api.app.api_v1.endpoints import ai_models

api_router = APIRouter()
# api_router.include_router(hello_world.router, prefix="/hello", tags=["hello"])
api_router.include_router(ai_models.router, prefix="/ai_models", tags=["AI Models"])


from typing import Type

from omnibridge.model_entities.base_model_entity import BaseModelUnit
from omnibridge.model_entities.models_io.base_model_io import ModelIO
from omnibridge.model_entities.models_io.base_model_io import TextualIO, ImageIO
from omnibridge.wrappers.wrapper_interfaces.model_wrapper import ModelWrapper


class TextToImageModel(BaseModelUnit):
    def __init__(self, textual_model_wrapper: ModelWrapper):
        self.wrapper = textual_model_wrapper

    def can_process_type(self, model_input_type: Type[ModelIO]) -> bool:
        return issubclass(model_input_type, TextualIO)

    def produced_type(self) -> Type[ImageIO]:
        return ImageIO

    def process(self, model_input: TextualIO) -> ImageIO:  # type: ignore
        if not self.can_process(model_input):
            raise Exception(f"Cannot process input of type: {type(model_input)}")

        prompt = model_input.get_text()
        response = self.wrapper.process(prompt=prompt)  # type: ignore
        image_output = ImageIO(response)  # type: ignore
        return image_output



import json
from typing import Dict, Any
from omnibridge.flows.flow_loader import FlowLoader
from omnibridge.flows.graph_flow import GraphFlow
from omnibridge.model_entities.models_io.base_model_io import TextualIO
from omnibridge.wrappers.wrapper_instances.type_name_to_wrapper import ModelLoader


def handle_run_command(args: Dict[str, Any]) -> None:
    object_to_run = None
    if args['object_to_run'] == 'model':
        object_to_run = ModelLoader.load_model(model_name=args['name'])
    elif args['object_to_run'] == 'flow':
        if args['file'] is not None:
            with open(args['file'], 'r') as f:
                data = json.load(f)
                object_to_run = GraphFlow.create_from_json("", data)
        else:
            object_to_run = FlowLoader.load_flow(args['name'])

    if not object_to_run:
        return
    
    response = object_to_run.process(TextualIO(args['prompt']))
    print(response)



import os

from omnibridge.model_entities.models_io.file_to_text_converter import FileInputHandler


def test_txt_file_conversion_to_textual_io():
    # Arrange
    content = 'Hello, World!'
    file_name = 'hello_for_test.txt'
    with open(file_name, 'w') as f:
        f.write(content)

    # Act
    text_io = FileInputHandler.convert_file(file_name)
    os.remove(file_name)  # cleanup

    # Assert
    assert text_io.get_text() == content


def test_py_file_conversion_to_textual_io():
    content = 'print("Hello, World!")'
    file_name = 'hello_for_test.py'
    with open(file_name, 'w') as f:
        f.write(content)

    # Act
    text_io = FileInputHandler.convert_file(file_name)
    os.remove(file_name)  # cleanup

    # Assert
    assert text_io.get_text() == content



from omnibridge.model_entities.models_io.base_model_io import TextualIO
from omnibridge.model_entities.models_io.flow_io import FlowIO


def test_flow_text_io_succeed():
    # Arrange
    flow_io = FlowIO(TextualIO("hello"))
    model_text_io = TextualIO("world")

    # Act
    flow_io += model_text_io

    # Assert
    assert flow_io.get_text() == "hello\nworld"




from abc import abstractmethod, ABC

from omnibridge.wrappers.wrapper_interfaces.model_wrapper import ModelWrapper


class Flow(ABC):
    @abstractmethod
    def add(self, new_model: ModelWrapper, instruction: str) -> None:
        pass



from typing import List, Dict, Any, Union

from omnibridge.model_entities.models_io.base_model_io import ModelIO, TextualIO
from omnibridge.model_entities.models_io.flow_io import FlowIO
from omnibridge.saved_data.json_data_manager import JsonConvertable
from omnibridge.wrappers.wrapper_instances.type_name_to_wrapper import ModelLoader
from omnibridge.wrappers.wrapper_interfaces.processor import Processor


class Node(JsonConvertable, Processor):
    def __init__(self, model: Processor, instruction: str = "", next_nodes: List[Processor] = []):
        self.next_nodes = next_nodes
        self.model = model
        self.instruction = instruction

    def process(self, model_input: ModelIO) -> FlowIO:
        assert isinstance(model_input, TextualIO)
        model_input = TextualIO(f"{model_input.get_text()} \n{self.instruction}")
        model_output = self.model.process(model_input)
        aggregate_output = FlowIO(model_output)
        for next_model in self.next_nodes:
            next_model_output = next_model.process(model_output)
            aggregate_output += next_model_output

        return aggregate_output

    def to_json(self) -> Dict[str, Union[str, int]]:
        raise NotImplementedError()

    @classmethod
    def create_from_json(cls, json_key: str, json_data: Dict[str, str]) -> Any:
        model_name = json_data['name']
        model = ModelLoader.load_model(model_name)

        instruction = json_data.get('instruction', "")

        next_models = []
        for next_model_data in json_data.get('models', []):
            next_model = Node.create_from_json("", next_model_data)  # type: ignore[arg-type]
            next_models.append(next_model)

        return Node(model, instruction, next_models)


class GraphFlow(JsonConvertable):
    def __init__(self, nodes: List[Node]):
        self.nodes = nodes

    def process(self, model_input: ModelIO) -> FlowIO:
        if len(self.nodes) == 0:
            raise ValueError("Cannot run flow with zero models!")

        flow_output = self.nodes[0].process(model_input)
        for node in self.nodes[1:]:
            flow_output += node.process(model_input)

        return flow_output

    def to_json(self) -> Dict[str, Union[str, int]]:
        raise NotImplementedError()

    @classmethod
    def create_from_json(cls, json_key: str, json_data: Dict[str, str]) -> Any:
        models_json = json_data['models']
        nodes = []
        for child_data in models_json:
            node = Node.create_from_json("", child_data)  # type: ignore[arg-type]
            nodes.append(node)
        return GraphFlow(nodes)



from typing import Dict, List, Union

from omnibridge.flows.flow import Flow
from omnibridge.model_entities.models_io.base_model_io import ModelIO, TextualIO
from omnibridge.model_entities.models_io.flow_io import FlowIO
from omnibridge.saved_data.json_data_manager import JsonConvertable
from omnibridge.wrappers.wrapper_instances.type_name_to_wrapper import ModelLoader
from omnibridge.wrappers.wrapper_interfaces.model_wrapper import ModelWrapper


class BranchingFlow(Flow, JsonConvertable):
    def __init__(self, name: str, root_model: ModelWrapper, branched_instructions: List[str],
                 branched_models: Union[List[ModelWrapper], None] = None):
        if branched_models is None:
            self.branched_models = [root_model for _ in range(len(branched_instructions))]
        elif len(branched_instructions) != len(branched_models):
            raise ValueError("Number of models and instructions must be the same.")
        else:
            self.branched_models = branched_models
        self.name = name
        self.root_model = root_model
        self.instructions = branched_instructions

    def process(self, model_input: ModelIO) -> FlowIO:
        root_output = self.root_model.process(model_input)
        assert isinstance(root_output, TextualIO)
        flow_output = FlowIO(root_output)
        for model, instruction in zip(self.branched_models, self.instructions):
            model_input = TextualIO(root_output.get_text() + "\n" + instruction)
            model_output = model.process(model_input)
            flow_output += model_output

        return flow_output

    def get_name(self) -> str:
        return self.name

    @classmethod
    def get_class_type_field(cls):
        return "branching"

    def to_json(self) -> Dict[str, Union[str, int]]:
        return {
            '_class_type': self.get_class_type_field(),
            'root_model': self.root_model.get_name(),
            'branched_models': ', '.join(model.get_name() for model in self.branched_models),
            'instructions': '$$$'.join(instruction for instruction in self.instructions)
        }

    @classmethod
    def create_from_json(cls, json_key: str, json_data: Dict[str, str]):
        name = json_key
        root_model = ModelLoader.load_model(json_data['root_model'])
        branched_models_names = json_data['branched_models'].split(', ')
        branched_models = [ModelLoader.load_model(model_name) for model_name in branched_models_names]
        branches_instruction = json_data['instructions'].split('$$$')
        return BranchingFlow(name, root_model, branches_instruction, branched_models)

    def add(self, new_model: ModelWrapper, instruction: str) -> None:
        self.branched_models.append(new_model)
        self.instructions.append(instruction)



from abc import abstractmethod

from omnibridge.saved_data.json_data_manager import JsonConvertable
from omnibridge.wrappers.wrapper_interfaces.processor import Processor


class ModelWrapper(JsonConvertable, Processor):
    @abstractmethod
    def get_name(self) -> str:
        pass

    @classmethod
    @abstractmethod
    def get_class_type_field(cls) -> str:
        pass



from typing import Dict, Type

from omnibridge.saved_data.json_data_manager import JsonDataManager, JsonConvertable
from omnibridge.wrappers.wrapper_instances.dalle_wrapper import DALLEWrapper
from omnibridge.wrappers.wrapper_instances.gpt_wrapper import GPTWrapper
from omnibridge.wrappers.wrapper_instances.hugging_face_wrappers import HuggingFaceWrapper
from omnibridge.wrappers.wrapper_interfaces.model_wrapper import ModelWrapper

type_names: Dict[str, Type[JsonConvertable]] = {
    GPTWrapper.get_class_type_field(): GPTWrapper,
    DALLEWrapper.get_class_type_field(): DALLEWrapper,
    HuggingFaceWrapper.get_class_type_field(): HuggingFaceWrapper
}


class ModelLoader:
    @staticmethod
    def load_model(model_name: str) -> ModelWrapper:
        class_type_name = JsonDataManager.get_json_value(["models", model_name, "_class_type"])
        wrapper_type = type_names[class_type_name]
        wrapper: ModelWrapper = JsonDataManager.load(["models", model_name], wrapper_type)
        return wrapper



from omnibridge.common.supported_models import SupportedModels
from typing import Any


def add_create_arguments(parser: Any) -> None:
    add_create_model_sub_parser(parser)
    add_create_flow_sub_parser(parser)
    add_create_key_sub_parser(parser)


def add_create_model_sub_parser(parser: Any) -> None:
    create_model_parser = parser.add_parser("model", help="Create a new model.")
    create_model_parser.add_argument('model_type', choices=[m.value for m in SupportedModels],
                                      help="The type of model to create.")
    create_model_parser.add_argument('-n', '--name', type=str, required=True, help="name of the model.")
    create_model_parser.add_argument('-k', '--key', type=str, required=True, help="api key name.")
    create_model_parser.add_argument('--sub-model', type=str,
                                      help="the sub model of the model youv'e chosen, for example" \
                                         "'gpt-3.5-turbo' if chatgpt is chosen, or 'repo_id' if huggingface is chosen")
    create_model_parser.add_argument('--num-images', type=str, default="4",
                                      help="number of images per prompt, default 4.")
    create_model_parser.add_argument('-r', '--res', type=str, default="256x256", help="resolution of images.")


def add_create_flow_sub_parser(parser: Any) -> None:
    create_flow_parser = parser.add_parser("flow", help="Create a new flow.")
    create_flow_parser.add_argument('-n', '--name', type=str, required=True, help="name of the flow.")
    create_flow_parser.add_argument('-t', '--type', type=str, choices=['seq', 'branching'], default='branching',
                                          help="type of the flow.")
    create_flow_parser.add_argument('-m', '--model', type=str,
                                          help="name of the model to use.")
    create_flow_parser.add_argument('-i', '--instruction', nargs='*', type=str,
                                          help="instruct the model to do something with the input.")
    create_flow_parser.add_argument('--multi', nargs='*', type=str,
                                          help="use multiple models. Currently only supported in sequential flow.")


def add_create_key_sub_parser(parser: Any) -> None:
    create_key_parser = parser.add_parser("key", help="Create an api key.")
    create_key_parser.add_argument('-n', '--name', help="name of the api key.", type=str, required=True)
    create_key_parser.add_argument('-v', '--value', help="value of the api key.", type=str, required=True)




from fastapi import APIRouter, status
from fastapi.responses import JSONResponse
from omnibridge_api.app.schemas.ai_model import AIModel, TextualModelPromptRequest, TextualModelPromptResponse
from pathlib import Path
from typing import List, Dict, Callable
import json

WRAPPER_TO_FUNC: Dict[str, Callable[[str, None], str]] = {} # TMP until cli implementation stabilizes

router = APIRouter()

@router.get("/models", response_model=List[AIModel], status_code=status.HTTP_200_OK)
def get_models() -> JSONResponse:
    models_path = Path(__file__).parents[2] / 'data/ai_models_metadata.json'

    try:
        with open(models_path, 'r') as f:
            models = json.load(f)
    except FileNotFoundError:
        return JSONResponse(status_code=status.HTTP_404_NOT_FOUND, content=[])

    return JSONResponse(status_code=status.HTTP_200_OK, content=models)


@router.post("/models/textual/prompts", response_model=TextualModelPromptResponse, status_code=status.HTTP_200_OK)
def prompt_textual_model(prompt_request: TextualModelPromptRequest) -> JSONResponse:
    model_name = prompt_request.model_name
    prompt = prompt_request.prompt

    wrapper_func = WRAPPER_TO_FUNC.get(model_name)

    if not wrapper_func:
        return JSONResponse(status_code=status.HTTP_501_NOT_IMPLEMENTED,
                             content={"response": f"{model_name} model is not supported"})


    return JSONResponse(status_code=status.HTTP_200_OK,
                         content={"response": wrapper_func(prompt, None)})



from typing import List

from omnibridge.model_entities.models_io.base_model_io import ModelIO, TextualIO, ImageIO


class FlowIO(ModelIO):
    def __init__(self, model_io: ModelIO):
        self.text = ""
        self.image_paths = []
        self.pretty_text = ""

        if isinstance(model_io, TextualIO):
            self.text = model_io.get_text()
            self.pretty_text = self.text

        elif isinstance(model_io, ImageIO):
            self.image_paths = model_io.image_paths()
            self.pretty_text = repr(model_io)

        else:
            raise NotImplementedError()

    def get_text(self) -> str:
        return self.text

    def get_pretty_text(self) -> str:
        return self.pretty_text

    def get_image_paths(self) -> List[str]:
        return self.image_paths

    def __repr__(self) -> str:
        return self.get_pretty_text()

    def __iadd__(self, other):  # type: ignore
        line_seperator = '\n\n' + ('*' * 100) + '\n\n'

        if isinstance(other, TextualIO):
            self.text += '\n' + other.get_text()
            self.pretty_text += line_seperator + other.get_text()
            return self

        elif isinstance(other, ImageIO):
            self.image_paths += other.image_paths()
            self.pretty_text += line_seperator + repr(other)
            return self

        elif isinstance(other, FlowIO):
            self.text += '\n' + other.get_text()
            self.image_paths += other.get_image_paths()
            self.pretty_text += line_seperator + other.get_pretty_text()
            return self

        else:
            raise TypeError(f"unsupported operand type(s) for +: '{type(self).__name__}' and '{type(other).__name__}'")




from typing import Dict, Any
from omnibridge.flows.branching_flow import BranchingFlow
from omnibridge.flows.sequential_flow import SequentialFlow
from omnibridge.wrappers.api_key import ApiKey
from omnibridge.saved_data.json_data_manager import JsonDataManager
from omnibridge.wrappers.wrapper_instances.gpt_wrapper import GPTWrapper
from omnibridge.wrappers.wrapper_instances.dalle_wrapper import DALLEWrapper
from omnibridge.wrappers.wrapper_instances.hugging_face_wrappers import HuggingFaceWrapper
from omnibridge.wrappers.wrapper_instances.type_name_to_wrapper import ModelLoader


def handle_create_command(args: Dict[str, Any]) -> None:
    if args['object_to_create'] == 'model':
        creation_function = MODEL_TYPE_TO_CREATION_FUNCTION.get(args['model_type'])
        if not creation_function:
            print('model type is not supported')
            return
        creation_function(args)

    elif args['object_to_create'] == 'flow':
        add_flow(args)

    elif args['object_to_create'] == 'key':
        add_key(args)


def add_key(args: Dict[str, Any]) -> None:
    api_key = ApiKey(args['value'])
    JsonDataManager.save(["api keys", args['name']], api_key)


def add_chatgpt(args: Dict[str, Any]) -> None:
    name = args['name']
    api_key: ApiKey = JsonDataManager.load(["api keys", args['key']], ApiKey)
    sub_model = args.get('sub_model')
    wrapper: GPTWrapper
    if not sub_model:
        wrapper = GPTWrapper(name, api_key.value)
    else:
        wrapper = GPTWrapper(name, api_key.value, sub_model)
    JsonDataManager.save(["models", name], wrapper)

def add_huggingface(args: Dict[str, Any]) -> None:
    name = args['name']
    api_key: ApiKey = JsonDataManager.load(["api keys", args['key']], ApiKey)
    wrapper: HuggingFaceWrapper = HuggingFaceWrapper(name, api_key.value, args['sub_model'])
    JsonDataManager.save(["models", name], wrapper)


def add_dalle(args: Dict[str, Any]) -> None:
    name = args['name']
    api_key: ApiKey = JsonDataManager.load(["api keys", args['key']], ApiKey)
    wrapper: DALLEWrapper = DALLEWrapper(name, api_key=api_key.value, number_of_images=args['num_images'],
                                            resolution=args['res'])
    JsonDataManager.save(["models", name], wrapper)


def add_flow(args: Dict[str, Any]) -> None:
    if args['type'] == 'branching':
            instructions = args['instruction']
            model = ModelLoader.load_model(args['model'])
            b_flow = BranchingFlow(args['name'], model, instructions)
            JsonDataManager.save(["flows", args["name"]], b_flow)
    elif args['type'] == 'seq':
        models = [ModelLoader.load_model(model_name) for model_name in args['multi']]
        flow = SequentialFlow(args['name'], models)
        JsonDataManager.save(["flows", args["name"]], flow)


MODEL_TYPE_TO_CREATION_FUNCTION = {
    'chatgpt': add_chatgpt,
    'dalle': add_dalle,
    'huggingface': add_huggingface,
}


from typing import Type, Dict

from omnibridge.flows.branching_flow import BranchingFlow
from omnibridge.flows.sequential_flow import SequentialFlow
from omnibridge.saved_data.json_data_manager import JsonDataManager, JsonConvertable
from omnibridge.wrappers.wrapper_interfaces.model_wrapper import ModelWrapper

flow_type_names: Dict[str, Type[JsonConvertable]] = {
    SequentialFlow.get_class_type_field(): SequentialFlow,
    BranchingFlow.get_class_type_field(): BranchingFlow,
}


class FlowLoader:
    @staticmethod
    def load_flow(flow_name: str) -> ModelWrapper:
        class_type_name = JsonDataManager.get_json_value(["flows", flow_name, "_class_type"])
        flow_type: Type[JsonConvertable] = flow_type_names[class_type_name]
        flow = JsonDataManager.load(["flows", flow_name], flow_type)
        return flow



import os

from omnibridge.flows.branching_flow import BranchingFlow
from omnibridge.flows.sequential_flow import SequentialFlow
import omnibridge.saved_data.json_data_manager
from omnibridge.saved_data.json_data_manager import JsonDataManager, MODULE_DIR
from omnibridge.wrappers.wrapper_instances.gpt_wrapper import GPTWrapper

TEST_FILE_NAME = ".saved_data_test.json"
TEST_FILE_PATH = os.path.join(MODULE_DIR, TEST_FILE_NAME)
omnibridge.saved_data.json_data_manager.FILE_PATH = TEST_FILE_PATH


def test_save_flow_succeed():
    # Arrange
    gpt = GPTWrapper("gpt3.5", api_key="fake_fake", model="gpt-3.5-turbo")
    JsonDataManager.save(["models", gpt.name], gpt)
    flow = SequentialFlow(name="myflow", models=[gpt])

    # Act
    JsonDataManager.save(["flows", flow.name], flow)
    loaded_flow: SequentialFlow = JsonDataManager.load(["flows", flow.name], SequentialFlow)

    # Assert
    model_original = flow.models[0]
    model_loaded = loaded_flow.models[0]
    assert isinstance(model_loaded, GPTWrapper)
    assert model_loaded.name == model_original.name
    assert model_loaded.model == model_original.model
    assert model_loaded.api_key == model_original.api_key


def test_save_flow_with_two_models_succeed():
    # Arrange
    gpt = GPTWrapper("gpt3.5", api_key="fake_fake", model="gpt-3.5-turbo")
    JsonDataManager.save(["models", gpt.name], gpt)
    flow = SequentialFlow(name="myflow", models=[gpt, gpt])

    # Act
    JsonDataManager.save(["flows", flow.name], flow)
    loaded_flow: SequentialFlow = JsonDataManager.load(["flows", flow.name], SequentialFlow)

    # Assert
    model_original = flow.models[0]
    model_original_2 = flow.models[1]
    model_loaded = loaded_flow.models[0]
    model_loaded_2 = loaded_flow.models[1]
    assert isinstance(model_loaded, GPTWrapper) and isinstance(model_loaded_2, GPTWrapper)
    assert model_loaded.name == model_original.name and model_loaded.model == model_original.model\
           and model_loaded.api_key == model_original.api_key
    assert model_loaded_2.name == model_original_2.name and model_loaded_2.model == model_original_2.model\
           and model_loaded_2.api_key == model_original_2.api_key


def test_save_branching_flow_succeed():
    # Arrange
    gpt = GPTWrapper("gpt3.5", api_key="fake_fake", model="gpt-3.5-turbo")
    JsonDataManager.save(["models", gpt.name], gpt)
    flow = BranchingFlow(name="myflow", root_model=gpt, branched_models=[gpt, gpt, gpt],
                         branched_instructions=["expand point 1", "expand point 2", "expand point 3"])

    # Act
    JsonDataManager.save(["flows", flow.name], flow)
    loaded_flow: BranchingFlow = JsonDataManager.load(["flows", flow.name], BranchingFlow)

    # Assert
    assert flow.instructions == loaded_flow.instructions



import pytest
from pathlib import Path
from omnibridge.cli.create.create_command_handler import add_key
import omnibridge.saved_data.json_data_manager


@pytest.fixture
def api_key_fixture():
    def _save_api_key_for_tests(file_path: Path, key_name: str):
        omnibridge.saved_data.json_data_manager.FILE_PATH = file_path
        args = {
            'value': 'api_key_mock',
            'name': key_name,
        }
        add_key(args)

    return _save_api_key_for_tests


from abc import abstractmethod, ABC

from omnibridge.model_entities.models_io.base_model_io import ModelIO


class Processor(ABC):
    @abstractmethod
    def process(self, model_input: ModelIO) -> ModelIO:
        pass



from typing import TypedDict, Any, Dict, Union
import json
import requests
import logging
from .rest_api_wrapper import RestAPIWrapper
from ...model_entities.models_io.base_model_io import ModelIO, TextualIO, ImageIO

IMAGE_GENARATION_API_URL = "https://api.openai.com/v1/images/generations"


class ImageGenerationRequestBody(TypedDict):
    prompt: str
    n: int
    size: str


class DALLEWrapper(RestAPIWrapper):
    def __init__(self, name: str, api_key: str, number_of_images: int, resolution: str,
                 logger: logging.Logger = logging.getLogger()) -> None:
        super().__init__(name, logger)
        self.api_url = IMAGE_GENARATION_API_URL
        self.api_key = api_key
        self.number_of_images = number_of_images
        self.resolution = resolution

    def to_json(self) -> Dict[str, Union[str, int]]:
        return {
            "api key": self.api_key,
            "number of images per prompt": self.number_of_images,
            "resolution": self.resolution,
            "_class_type": self.get_class_type_field()
        }

    @classmethod
    def create_from_json(cls, json_key: str, json_data: Dict[str, str]) -> Any:
        return DALLEWrapper(json_key, api_key=json_data["api key"],
                            number_of_images=int(json_data["number of images per prompt"]),
                            resolution=json_data["resolution"])

    @classmethod
    def get_description(cls) -> str:
        return """
            DALLE-2 OpenAI Wrapper, generates an image based on text
        """

    @classmethod
    def get_class_type_field(cls) -> str:
        return "dalle"

    def _get_body(self, prompt_message: str) -> Any:
        return json.dumps({
            "prompt": prompt_message,
            "n": self.number_of_images,
            "size": self.resolution
        })

    def _get_api_url(self) -> str:
        return self.api_url

    def _get_api_key(self) -> str:
        return self.api_key

    def process(self, model_input: ModelIO) -> ImageIO:
        if not isinstance(model_input, TextualIO):
            raise TypeError(f"expected type {type(TextualIO)} but got {type(model_input)}")
        response = self.prompt(model_input.text)

        images = []
        images_path = []
        data = response['data']

        for idx, item in enumerate(data):
            response = requests.get(item['url'])
            images.append(response.content)

            with open(f"image_{idx}.jpg", "wb") as f:
                f.write(response.content)
            images_path.append(f"image_{idx}.jpg")

        return ImageIO(images_path)



import os
from typing import Dict

import pytest

import omnibridge.saved_data.json_data_manager
from omnibridge.saved_data.json_data_manager import JsonConvertable, JsonDataManager, MODULE_DIR


TEST_FILE_NAME = ".saved_data_test.json"
TEST_FILE_PATH = os.path.join(MODULE_DIR, TEST_FILE_NAME)
omnibridge.saved_data.json_data_manager.FILE_PATH = TEST_FILE_PATH


class Person(JsonConvertable):
    def __init__(self, name: str, age: str):
        self.name = name
        self.age = age

    def to_json(self) -> Dict[str, str]:
        return {
            "name": self.name,
            "age": self.age
        }

    @classmethod
    def create_from_json(cls, json_key: str, json_data: Dict[str, str]):
        return Person(json_data['name'], json_data['age'])


def test_save_and_load_json_succeed():
    # Arrange
    my_person = Person(name="jack", age="45")

    # Act
    JsonDataManager.save(["my_person"], my_person)
    created_person = JsonDataManager.load(["my_person"], Person)

    # Assert
    assert my_person.age == created_person.age and my_person.name == created_person.name


def test_load_json_file_not_exist_fail():
    # Arrange
    if os.path.isfile(TEST_FILE_PATH):
        os.remove(TEST_FILE_PATH)

    # Act + Assert
    with pytest.raises(FileNotFoundError):
        JsonDataManager.load(["my_person"], Person)


def test_load_json_key_not_exist_fail():
    # Arrange
    if os.path.isfile(TEST_FILE_PATH):
        os.remove(TEST_FILE_PATH)

    person = Person("Bruce", "50")

    # Act + Assert
    JsonDataManager.save(["bruce"], person)

    with pytest.raises(KeyError):
        JsonDataManager.load(["wayne"], Person)


def test_save_and_load_multiple_items_json_succeed():
    # Arrange
    person1 = Person(name="jack", age="45")
    person2 = Person(name="Li", age="32")

    # Act
    JsonDataManager.save([person1.name], person1)
    JsonDataManager.save([person2.name], person2)
    created_person1 = JsonDataManager.load([person1.name], Person)
    created_person2 = JsonDataManager.load([person2.name], Person)

    # Assert
    assert created_person1.age == person1.age and created_person1.name == person1.name
    assert created_person2.age == person2.age and created_person2.name == person2.name


def test_save_and_load_multiple_nested_items_json_succeed():
    # Arrange
    person1 = Person(name="jack", age="45")
    person2 = Person(name="Li", age="32")

    # Act
    JsonDataManager.save(["employees", person1.name], person1)
    JsonDataManager.save(["managers", person2.name], person2)
    created_person1 = JsonDataManager.load(["employees", person1.name], Person)
    created_person2 = JsonDataManager.load(["managers", person2.name], Person)

    # Assert
    assert created_person1.age == person1.age and created_person1.name == person1.name
    assert created_person2.age == person2.age and created_person2.name == person2.name


def test_overwrite_json_succeed():
    old_person = Person(name="wrongly typed", age="")
    updated_person = Person(name="Gal", age="28")

    # Act
    JsonDataManager.save(["my_person"], old_person)
    JsonDataManager.save(["my_person"], updated_person)
    created_person = JsonDataManager.load(["my_person"], Person)

    # Assert
    assert updated_person.age == created_person.age and updated_person.name == created_person.name


def test_save_multiple_classes_json_succeed():
    # Arrange
    class Pet(JsonConvertable):
        def __init__(self, nickname: str):
            self.nickname = nickname

        def to_json(self) -> Dict[str, str]:
            return {
                'nickname': self.nickname
            }

        @classmethod
        def create_from_json(cls, json_key: str, json_data: Dict[str, str]):
            return Pet(json_data['nickname'])

    person = Person("owner", "32")
    pet = Pet("dogo")

    # Act
    JsonDataManager.save(["my_person"], person)
    JsonDataManager.save(["my_pet"], pet)
    loaded_person = JsonDataManager.load(["my_person"], Person)
    loaded_pet = JsonDataManager.load(["my_pet"], Pet)

    # Assert
    assert loaded_person.name == person.name and loaded_pet.nickname == pet.nickname



from typing import Any


def add_run_arguments(parser: Any) -> None:
    add_run_flow_sub_parser(parser)
    add_run_model_sub_parser(parser)


def add_run_flow_sub_parser(parser: Any) -> None:
    run_flow_parser = parser.add_parser("flow", help="run flow.")
    group = run_flow_parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-n', '--name', type=str, help="name of the flow.")
    group.add_argument('-f', '--file', type=str, help="name of json file containing a flow definition.")
    run_flow_parser.add_argument('-p', '--prompt', help="prompt for flow", type=str, required=True)


def add_run_model_sub_parser(parser: Any) -> None:
    run_flow_parser = parser.add_parser("model", help="run flow.")
    run_flow_parser.add_argument('-n', '--name', type=str, required=True, help="name of the flow.")
    run_flow_parser.add_argument('-p', '--prompt', help="prompt for flow", type=str, required=True)


import responses
import pytest

from omnibridge.model_entities.models_io.base_model_io import TextualIO
from omnibridge.wrappers.wrapper_instances.rest_api_wrapper import WrapperException
from omnibridge.wrappers.wrapper_instances.gpt_wrapper import COMPLETIONS_API_URL, GPTWrapper


@pytest.fixture
def well_structured_response():
    return {
        "choices": [
            {
                "message": {
                    "content": "mock"
                }
            }
        ]
    }


@pytest.fixture
def bad_structured_response():
    return {
        "bad-structure": [
            {
                "mock": "mock"
            }
        ]
    }


@responses.activate
def test_gpt_wrapper(well_structured_response):
    # Arrange
    responses.add(responses.POST, COMPLETIONS_API_URL, json=well_structured_response)
    wrapper = GPTWrapper(name='testgpt', api_key='abc', model='gpt-4')

    # Act
    res = wrapper.process(TextualIO('send_mock'))

    # Assert
    assert len(responses.calls) == 1
    assert responses.calls[0].request.url == COMPLETIONS_API_URL
    assert res.text == 'mock'


@responses.activate
def test_gpt_wrapper_fail(bad_structured_response):
    # Arrange
    responses.add(responses.POST, COMPLETIONS_API_URL, json=bad_structured_response)
    wrapper = GPTWrapper(name='testgpt', api_key='abc', model='gpt-4')

    # Act
    with pytest.raises(KeyError) as exc_info:
        wrapper.process(TextualIO('send_mock'))

    # Assert
    assert exc_info.value.args[0] == "choices"


@responses.activate
def test_gpt_wrapper_api_call_fails():
    # Arrange
    responses.add(responses.POST, COMPLETIONS_API_URL, json={}, status=500)
    wrapper = GPTWrapper(name='testgpt', api_key='abc', model='gpt-4')

    # Act
    with pytest.raises(WrapperException) as exc_info:
        wrapper.process(TextualIO('send_mock'))

    # Assert
    assert str(exc_info.value).startswith(f"Request to api endpoint: {COMPLETIONS_API_URL} failed.")



from abc import ABC
from typing import List


class ModelIO(ABC):
    pass


class TextualIO(ModelIO):
    def __init__(self, text: str):
        self.text = text

    def get_text(self) -> str:
        return self.text

    def __repr__(self):  # type: ignore
        return self.get_text()


class ImageIO(ModelIO):
    def __init__(self, paths: List[str]):
        self.paths = paths

    def image_paths(self):  # type: ignore
        return self.paths

    def __repr__(self):  # type: ignore
        return f"image paths: {self.paths}"


class FileIO(ModelIO):
    pass



from typing import Dict, Any
from omnibridge.wrappers.wrapper_instances.type_name_to_wrapper import type_names
from omnibridge.saved_data.json_data_manager import JsonDataManager
import json


def handle_list_command(args: Dict[str, Any]) -> None:
    func = OBJECT_TO_LIST_FUNCTION.get(args['object_to_list'])

    if not func:
        return
    
    func()


def list_wrappers() -> None:
    print([type for type in type_names.keys()])


def list_keys() -> None:
    try:
        keys = JsonDataManager.get_json_value(['api keys'])
        print(json.dumps(keys, indent=2))
    except KeyError:
        print('No keys were found in your saved data')
    

def list_models() -> None:
    try:
        models = JsonDataManager.get_json_value(['models'])
        print(json.dumps(models, indent=2))
    except KeyError:
        print('No models were found in your saved data')



def list_flows() -> None:
    try:
        flows = JsonDataManager.get_json_value(['flows'])
        print(json.dumps(flows, indent=2))
    except KeyError:
        print('No flows were found in your saved data')
   


OBJECT_TO_LIST_FUNCTION = {
    'wrappers': list_wrappers,
    'keys': list_keys,
    'models': list_models,
    'flows': list_flows
}


from pathlib import Path
from omnibridge.cli.create.create_command_handler import add_key, add_chatgpt, add_dalle
import omnibridge.saved_data.json_data_manager

import json



def test_add_key_command(tmp_path: Path, monkeypatch):
    # Arrange
    file_path = tmp_path / ".saved_data.json"
    monkeypatch.setattr(omnibridge.saved_data.json_data_manager, "FILE_PATH", file_path)
    args = {
        'value': 'value_mock',
        'name': 'name_mock'
    }

    # Act
    add_key(args)

    with open(file_path, 'r') as f:
        saved_data = json.load(f)


    # Assert
    assert saved_data == {
        'api keys': {
            'name_mock': {
                'value': 'value_mock'
            }
        }
    }


def test_add_chatgpt_command(api_key_fixture, tmp_path: Path, monkeypatch):
    # Arrange
    file_path = tmp_path / ".saved_data.json"
    monkeypatch.setattr(omnibridge.saved_data.json_data_manager, "FILE_PATH", file_path)
    key_name = "mock_key_name"
    model_name = 'gpt_model'
    model = 1
    api_key_fixture(file_path, key_name)

    args = {
        'model': model,
        'name': model_name,
        'key': key_name,
        'sub_model': 1
    }


    # Act
    add_chatgpt(args)

    with open(file_path, 'r') as f:
        saved_data = json.load(f)


    # Assert
    assert saved_data == {
        'api keys': {
            key_name:  {
                'value': 'api_key_mock'
            }
        },
        'models': {
            model_name: {
                '_class_type': 'chat_gpt',
                'api key': 'api_key_mock',
                'model': model
            }
        }
    }

    
def test_add_dalle_command(api_key_fixture, tmp_path: Path, monkeypatch):
    # Arrange
    file_path = tmp_path / ".saved_data.json"
    monkeypatch.setattr(omnibridge.saved_data.json_data_manager, "FILE_PATH", file_path)
    key_name = "mock_key_name"
    model_name = 'dalle_model'
    num_of_images = 4
    resolution = 'mock_resolution'
    api_key_fixture(file_path, key_name)

    args = {
        'key': key_name,
        'num_images': num_of_images,
        'res': resolution,
        'name': model_name
    }

    # Act
    add_dalle(args)

    with open(file_path, 'r') as f:
        saved_data = json.load(f)

    
    # Assert
    assert saved_data == {
        'api keys': {
            key_name:  {
                'value': 'api_key_mock'
            }
        },
        'models': {
            'dalle_model': {
                '_class_type': 'dalle',
                'api key': 'api_key_mock',
                'number of images per prompt': num_of_images,
                'resolution': resolution
            }
        }
    }


from omnibridge.flows.graph_flow import Node, GraphFlow
from omnibridge.model_entities.models_io.base_model_io import ModelIO, TextualIO, ImageIO
from omnibridge.wrappers.wrapper_interfaces.processor import Processor


class MockProcessor(Processor):
    processed = 0
    output = "test_output"

    def __init__(self, return_value: ModelIO = TextualIO(output)):
        self.return_value = return_value

    def process(self, model_input: ModelIO) -> ModelIO:
        MockProcessor.processed += 1
        return self.return_value


def test_single_node_graph_flow():
    # Arrange
    MockProcessor.processed = 0
    node = Node(MockProcessor())
    flow = GraphFlow([node])
    flow_input = TextualIO("test_input")

    # Act
    flow.process(flow_input)

    # Assert
    assert MockProcessor.processed == 1


def test_chained_node_graph_flow():
    # Arrange
    MockProcessor.processed = 0
    last_node = Node(MockProcessor())
    node = Node(MockProcessor(), "", [last_node])
    flow = GraphFlow([node])
    flow_input = TextualIO("test_input")

    # Act
    flow.process(flow_input)

    # Assert
    assert MockProcessor.processed == 2


def test_parallel_node_graph_flow():
    # Arrange
    MockProcessor.processed = 0
    node = Node(MockProcessor())
    flow = GraphFlow([node, node])
    flow_input = TextualIO("test_input")

    # Act
    flow.process(flow_input)

    # Assert
    assert MockProcessor.processed == 2


def test_complex_graph_flow():
    # Arrange
    MockProcessor.processed = 0
    last_nodes = Node(MockProcessor(ImageIO(["./file1", "./file2"])))
    nodes = [Node(MockProcessor(), instruction="", next_nodes=[last_nodes]) for _ in range(2)]
    flow = GraphFlow(nodes)
    flow_input = TextualIO("test_input")

    # Act
    flow.process(flow_input)

    # Assert
    assert MockProcessor.processed == 4



from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from omnibridge_api.app.core.config import settings
from omnibridge_api.app.api_v1.api import api_router

origins = [
    "http://localhost:3000",
    "localhost:3000"
]

app = FastAPI(
    title=settings.PROJECT_NAME, openapi_url=f"{settings.API_V1_STR}/openapi.json"
)

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"]
)


app.include_router(api_router, prefix=settings.API_V1_STR)
